<%* 
/*  Load all super templates
	1. Read all super templates from custom folder set below
	2. Load all variables from the super-templates into a super array 
	3. Each member of the new array will contain two arrays, "common" and "main". "common" contains properties in common to one superclass; "main" is an array of objects or dictionaries that contains subclasses of properties which share a superclass.
	4. The purpose of the super-templates is reducing the number of templater files and make it easier to maintain from one central location.
*/
const templatesFolder = ['Obsidian/02-Templater'];
const test = false;
const templateFiles = []   // names of super template files
for (const folder of templatesFolder) {
  const files = (await app.vault.adapter.list(folder))?.files || []
  files.sort((a, b) => a.localeCompare(b)) // alphabetically
  templateFiles.push(...files) }
if (!templateFiles) return   // no super template found
const templates = templateFiles.map(path => { // read base names
	const file = app.vault.getAbstractFileByPath(path)
	return file.basename })
const superTemplates = [];   // array will contain all super-templates
for (item in templates) {    // iterate through each super-template
	superTemplates.push(await tp.file.include(`[[${templates[item]}]]`))
} // super template variables loaded in super array
let title = tp.file.title;   // grab current title
if (title.startsWith("Untitled")) {
	/** Create a note through multiple suggesters. Two options here:
		1. A new note prompt to type in the name of the note. This title may or may not contain a prefix. Call "newNoteFromPrefix"
		2. More templates options where the subclasses in the super-templates are exposed. Call "newNoteThruSelectors"
	 */
	const options = ["New Note", "More Templates"];
	const choice = await tp.system.suggester(options, options)
	switch(choice) {
		case "New Note":
			title = await tp.system.prompt("Note Name")
			if (title) {
				const output = await newNoteFromPrefix(superTemplates) 
				if (output) {
					tR += output;
					break; }
			}
			tR += await runDefaultMinimal()
			break;
		case "More Templates":    // show a templater suggester
			const output = await newNoteThruSelectors(superTemplates);	
			if (output) {
				tR += output;
				break; } 
			tR += await runDefaultMinimal();
			break;
		default:   // nothing selected
			tR += await runDefaultMinimal();
	}
} else {   
	/**
		Note to create is a link that has a prefix and a title; 
		name is not "Untitled"
	 */
	tR += await newNoteFromPrefix(superTemplates)
}

/**
    Run the Minimal Default template for the cases when user presses Escape, Enter, or cancels a selection
 */
async function runDefaultMinimal() {
	await new Promise((resolve, reject) => setTimeout(resolve, 100));
	return await tp.file.include("[[Default Minimal Template]]");
}

/**
 * CREATING A NEW NOTE THAT IS AN EXISTING LINK WITH A PREFIX
   The user clicked on a link-title typed in a note. Process according to prefix of link-title in a note. This assumes we are starting with a note not named "Untitled" because the note is receiving its title from the link.
 * @param {array} sT - an array with all super-templates
 * return {string} frontmatter
 */
async function newNoteFromPrefix(sT) {
	const foundClass = findPrefixInSuperArray(sT);
	if (foundClass) {
		console.log("Found prefix in title: ", title)
		const subCommon = foundClass.common;
		const subClass  = foundClass.main;
		title = extractTitle(title);  // get rid of the prefix
		await tp.file.rename(title);
		return processFrontmatter(subCommon, subClass);
	} else { // prefix not found in super-array
		console.log("No prefix in title: ", title)
		setTimeout(() => { tp.file.rename(title) }, 300)
		return await tp.file.include("[[Default Minimal Template]]");
	}  
}

/**
 * CREATING A NEW NOTE USING A NESTED SUGGESTER
 * Nested suggester for all super-templates. The user is creating a note from scratch by pressing Ctrl N. By default, the name of the new note is “Untitled”. 
 * @param {array} sT - an array with all super-templates
 * return {string} frontmatter
 */
async function newNoteThruSelectors(sT) {
	const supClasses = getSuperClasses(sT);
	const SClass = await tp.system.suggester(supClasses, sT)
	if (!SClass) return
	// wait a bit for the suggester to return the array
	await new Promise((resolve, reject) => setTimeout(resolve, 300));
	const classes = getValuesToListGivenKey(SClass.main, "name") 
	const subClass = await tp.system.suggester(classes, SClass.main)
	if (!subClass) return;             // nothing selected
	const subCommon = SClass.common;   // load common properties
	return processFrontmatter(subCommon, subClass);
}

/**
 * Build the frontmatter from the selected arrays and dictionaries
 * @param {array} subCommon - array for subclass common properties
 * @param {array} subClass - array of subclass main properties
 * return {string} frontmatter
*/
async function processFrontmatter(subCommon, subClass) {
	const properties = cleanFrontmatter(subClass.fields)
	let topProps = getTopProperties(subCommon)    // read top properties
	let botProps = getBottomProperties(subCommon) // bottom props
	if (subClass.exclude) {   // there is a list of exclusions
		topProps = removeExcludedKeys(topProps, subClass.exclude)
		botProps = removeExcludedKeys(botProps, subClass.exclude)
	}   // properties in exclude removed
	botProps = replace_uuid(botProps);
	const dataview = subClass.dataview   // add function if exists
			? addTableBelowFrontmatter(subClass.dataview)
			: ''
	const folder = subClass.folder ? subClass.folder + "/" : "/"; 
	await tp.file.move(folder + title);   // move to class sub-folder
	return addFrontmatter(topProps + properties + botProps) + dataview	
}

/**
 * Get the name of the super classes for all super-templates. 
 * The class name in in the array "common"
 * @param {array} sT - array of super template objects
 * return {array} name of super classes
 */
function getSuperClasses(sT) {
	let result = [];
	sT.forEach(i => {
		result.push(i.common[2].class) })
	return result;
}

/**
 * Make an uuid to time stamp each new note.  If true, return frontmatter as "uuid: '20240220103600'"; if false, only the string 20240220103600
 * @param {boolean} withProperty 
 * return {string} an uuid string or the uuid property
 */
function get_uuid(withProperty) {
	return withProperty 
			? "uuid: " + "'" 
				+  tp.file.creation_date("YYYYMMDDHHmmss")  + "'"
			: tp.file.creation_date("YYYYMMDDHHmmss")
}

/**
 * Replace the updated uuid in the current properties
 * @param {string} a YAML like string
 * return {string} a string with an updated uuid
 */
function replace_uuid(txt) {
	return txt + get_uuid(true)
}

/**
 * Return an array of values for a given dictionary and key. Important for providing the names and prefixes of property subclasses and validate an entry
 * @param {array} arr - array of objects in a superclass 
 * @param {string} key - a key in a dictionary
 * return {array} an array of values corresponding to a key
 */
function getValuesToListGivenKey(arr, key) {
	let result = [];
	arr.forEach(item => {
		let dict = item;
		let value = dict[key];
		// console.log(`${key}: ${value}`);
		result.push(value) })
	return result;
}

/**
 * Remove blank lines, spaces and tabs at the start of frontmatter caused by the indentations
 * @param {string} yml - a YAML like text
 * return {string} without blank lines
**/
function cleanFrontmatter(yml) {
	const regex = /^(?=\n)$|^\s*|s*$|\n\n+/gm   // line breaks, spaces
	yml = yml.split('\n')
		.map(function(line) {
			return line.replace(regex, "")})
		.filter(function(x) {return x}).join("\n");
	return yml + "\n";   // adding blank line so no mixed with next
}

/**
 * Get the common properties, top or bottom depending of the index
 * @param {array} arr - the common array from a super-template
 * @param {integer} idx - a number 0 or 1. 0 for top; 1 for bottom
 * return {array} common properties
 */
function getCommonProperties(arr, idx) {
	return arr[idx].fields
}

/**
 * Get the common top properties
 * @param {array} arr - the common array from a super-template
 * return {string} a string with the top common properties 
 */
function getTopProperties(arr) {
	return cleanFrontmatter(getCommonProperties(arr, 0))
}

/**
 * Get the common bottom properties
 * @param {array} arr - the common array from a super-template
 * return {string} a string with the bottom common properties 
 */
function getBottomProperties(arr) {
	return cleanFrontmatter(getCommonProperties(arr, 1))
}

/**
 * Print the frontmatter to the note
 * @param {string} yml - text with properties
 * return {void}
 */
function printFrontmatter(yml) {
	tR += "---\n"
	tR += yml
	tR += "---\n"
}

/**
 * Read the field prefix from each dictionary in the array
 * @param {array} arr - array of objects
 * return {array} list of all prefixes in super-templates
 */
function getPrefixValuesToList(arr) {
	const key = "prefix";
	return getValuesToListGivenKey(arr, key)
}

/**
 * Find the prefix that was entered in the note title through all the super-templates. It will iterate through each array in the main super array and will stop when finding a prefix match.
 * @param {array} superArr - array of super-templates
 * return {object} matched dictionary 
 */
function findPrefixInSuperArray(superArr) {
	console.log(superArr)
	let choice;
	let main;
	let common;
	superArr.forEach(item => {
		item.main.map(d => {
			if (test) console.log(d.prefix);
			if (title.startsWith(d.prefix)) {
				choice = item;
				main = d
			 }
		})
	})
	return !common && !main 
		? undefined
		: { common: choice.common, main: main };
}

/**
 * Clean the YAML string and remove the excluded properties
 * @param {string} txt - a string of properties or fields
 * @param {array} exclude - a list with the properties to exclude
 * return {string} a clean text with some properties removed
 */
function cleanRemovedExcluded(txt, exclude) {
	return cleanFrontmatter(removeExcludedKeys(txt, exclude));
}

/**
 * Remove properties specified in the array exclude because some classes do not need all the common properties.
 * @param {string} txt - original fields
 * @param {array} exclude - list of exclusions given by "main"
 * return {string} removed properties
 */
function removeExcludedKeys(txt, exclude) {
	removeList = exclude.map(p => p+":")
	var expStr = removeList.join("|");
	return txt
		.replace(new RegExp(expStr, 'gi'), ' ')
		.replace(/\s{2,}/g, '\n')  // remove void lines after exclusion
}

/**
 * Add dashes to properties to form the YAML frontmatter
 * @param {string} yml - a string of properties or fields
 * return {string}  frontmatter
 */
function addFrontmatter(yml) {
	return "---\n" + yml + "\n---\n"
}

/**
 * Print YAML frontmatter after adding dashes to properties
 * @param {string} yml - a string of properties or fields
 * return {void}
 */
function printFrontmatter(yml) {
	tR += "---\n"
	tR += yml
	tR += "\n---\n"
}

/**
 * Extract the title after the prefix. Remember that the dash is the delimiter for "prefix-title". Date regex to detect if there is a date in the title. We don't want to split it at yyyy- or mm-. Additionally, title could be a date "2024-02-21", or of the form "prefix-Battery-powered-device"
 * @param {string} title - note title with a prefix and extra dashes
 * return {string} just the note title
 */
function extractTitle(title) {
	let result;
	let count = (title.match(/-/g) || []).length;
	let dateRegex = /(\d{4})([-])(\d{2})([-])(\d{2})/;
	
	switch (count) {
		case 0:      // title has no dash, no prefix
			result = title.trim()
			break
		case 1:      // title has one dash, means prefix
			result = title.split("-").slice(1)[0].trim()
			break
		default:     // more than one dash
			result = dateRegex.test(title) && count == 2
				? title.trim() 
				: title.split("-").slice(1).join("-").trim()
	}
	return result;
}

/**
 * Insert a Yaml like string along the dataview query
 * @param {string} fun - a function as a string
 * return {string} text with a dataview block including the resolved function with DQL code
 */
function addTableBelowFrontmatter(fun) {
	return "\n\n```dataview\n" + cleanFrontmatter(eval(fun))
	+ "```"
}

/**
 * Returns a dataview table showing notes where their properties up and related contain a link to the title of the current note. There could be as many as these DV functions as needed. But they have to be entered as strings in their respective dictionary.
 * @param {void} 
 * return {string} the Dataview Query as text 
 */
function getTableUpRelated() {
	return `
	TABLE up, related
	FROM ""
	WHERE up = link(this.file.name)
	   OR contains(related, link(this.file.name))
	SORT file.name ASC
	`
}
-%>